from typing import Dict, Any
import datetime
import logging

from amarium.utils import prepare_file_name_saving


class MetricsLogger:
    """The class implements the Logger factory for basic metrics used in ML and
    deep Learning 
    
    :author: Julian M. Kleber
    """

    def __init__(self, save_dir: str) -> None:

        self.logger = logging.getLogger("metrics_logger")
        self.time_initialized = datetime.datetime.now().strftime("%Y-%m-%d_%H:%M:%S")
        self.filename = prepare_file_name_saving(
            prefix=save_dir, file_name=f"{self.time_initialized}-metrics_logger", suffix=".log"
        )
        
        self.formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        self.fileHandler = logging.FileHandler(self.filename)
        self.encoding = "utf-8"

        
        self.logger.setLevel(logging.DEBUG)

        self.fileHandler.setFormatter(self.formatter)
        self.logger.addHandler(self.fileHandler)

        self.logger.info(f"Initialized Metrics Logger to {self.filename}")

    def log(self, metrics: Dict[str, Any], level: str = "info") -> None:
        """Logs the metrics using the self.logger."""
        for key, val in metrics.items():

            if level == "info":
                self.logger.info(self.basic_layout(key, val))
            elif level == "debug":
                self.logger.info(self.basic_layout(key, val))
            elif level == "warning":
                self.logger.warning(self.basic_layout(key, val))
            elif level == "error":
                self.logger.error(self.basic_layout(key, val))
            elif level == "critical":
                self.logger.critical(self.basic_layout(key, val))
            else:
                self.logger.warning(
                    f"Did not specify correct log level with {level}. Please either specify "
                    "debug, info, warning, error or critical"
                )
                self.logger.info(self.basic_layout(key, val))

    def basic_layout(self, metric: str, value: str) -> str:
        """
        Defines the basic logging layout for the MetricsLogger
        
        :param: metric: str: Name of the metric
        :paraM: value: str: Value of the metric 
        :author: Julian M. Kleber
        """
        return f"{metric} : {value}"
    
    def close_logger(self)->None: 
        """
        To limit logging to one file for multi run experiments
        the function closes the current logging file. 

        :author: Julian M. Kleber
        """
        handlers = self.logger.handlers
        for handler in handlers: 
            if isinstance(handler, logging.FileHandler):
                self.logger.removeHandler(handler)
                handler.close()



import torch
import torch.nn.functional as F
from torch.nn import Linear

from torch_geometric.nn import global_add_pool, GraphConv, GATConv

import numpy as np
import sklearn.metrics as metrics

import logging


from carate.models.base_model import Model

logger = logging.getLogger(__name__)
logging.basicConfig(
    filename="carate.log",
    encoding="utf-8",
    level=logging.DEBUG,
    format="%(asctime)s %(message)s",
)


class Net(Model):
    """
    The Net is the core algorithm and needs a constructor and a
    forward pass. The train, test and evaluation methods are implemented
    in the evaluation module with the Evaluation class.
    """

    def __init__(
        self,
        dim: int,
        num_features: int,
        num_classes: int,
        num_heads: int = 16,
        dropout_gat: float = 0.6,
        dropout_forward: float = 0.5,
        *args,
        **kwargs,
    ) -> None:
        super(Net, self).__init__(
            dim=dim, num_classes=num_classes, num_features=num_features
        )

        self.num_heads = num_heads
        self.dropout_gat = dropout_gat
        self.dropout_forward = dropout_forward

        self.dropout_forward = dropout_forward

        self.conv1 = GATConv(
            self.num_features, self.dim, dropout=self.dropout_gat, heads=self.num_heads
        )

        self.fc1 = Linear(self.dim * self.num_heads, self.dim)
        self.fc2 = Linear(self.dim, self.num_classes)

    def forward(self, x, edge_index, batch, edge_weight=None):
        x = F.relu(self.conv1(x, edge_index, edge_weight))
        x = F.dropout(x, p=self.dropout_forward, training=self.training)
        x = global_add_pool(x, batch)

        x = F.relu(self.fc1(x))
        x = self.fc2(x)
        return x
    def __str__(self): 
        return "g_regression"


import os 

from carate.logging.metrics_logger import MetricsLogger
from amarium.utils import delete_file, delete_all_files_in_dir, delete_empty_dir, make_full_filename

def test_metrics_logger():

    if os.path.isdir("tests/test_loggers/"):
        delete_all_files_in_dir("tests/test_loggers/")

    save_dir = "tests/test_loggers/"

    metricsLogger = MetricsLogger(save_dir)
    metricsLogger.log({"MAE":0.001}, "info")
    assert os.path.basename(metricsLogger.filename) in os.listdir("tests/test_loggers")

    delete_all_files_in_dir("tests/test_loggers/")
    delete_empty_dir("tests/test_loggers/")

    assert not os.path.isdir("test/test_loggers/")





"""
Module to perform analysis of runs. 

:author: Julian M. Kleber
"""
from typing import List, Dict, Any, Tuple
import numpy as np
from amarium.utils import attach_slash, load_json_from_file
import logging

logging.basicConfig(
    filename="carate.log",
    encoding="utf-8",
    level=logging.DEBUG,
    format="%(asctime)s %(message)s",
)


def get_best_average(result_list: List[float], key_val: str) -> List[float]:
    """
    The get_best_average function takes a list of floats and returns the average of the best values
    in that list.

    Returns: List[float]: A list containing only the best averages for each run in step_list.

    :param step_list:List[float]: Used to Determine the step size for each run.
    :param key_val:str: Used to Specify the key value of the dictionary that is being unpacked.
    :return: The best average reward for each run.

    :doc-author: Trelent
    """

    step_list = unpack_run(result_list, key_val)
    best_vals = np.max(step_list, axis=1)
    avg = np.mean(best_vals)
    std = np.std(best_vals)
    print(avg, "pm", std)
    return []


def get_avg(step_list: List[float]) -> float:
    """
    The get_avg function takes a list of floats and returns the average value.


    :param step_list:List[float]: Used to Tell the function that it will be taking a list of
    floats as an argument.
    :return: The mean of the step_list.

    :doc-author: Julian M. Kleber
    """

    return float(np.mean(step_list))


def get_max(step_list: List[float]) -> float:
    """
    The get_max function takes a list of floats and returns the maximum value in that list.


    :param step_list:List[float]: Used to Tell the function that step_list is a list of floats.
    :return: The maximum value in the list.

    :doc-author: Julian M. Kleber
    """

    return float(np.max(step_list))


def get_min(step_list: List[float]) -> float:
    """
    The get_min function takes a list of floats and returns the minimum value in that list.

    :param step_list:List[float]: Used to Specify the type of parameter that is being passed
    into the function.
    :return: The minimum value in the step_list.

    :doc-author: Julian M. Kleber
    """

    return float(np.min(step_list))


def get_min_max_avg_cv_run(
    result: List[Dict[str, List[float]]], key_val: str
) -> Tuple[List[float]]:
    """
    The get_min_max_avg_cv_run function takes in a list of dictionaries, and a key value.
    It then unpacks the values associated with that key into three lists: max_val, min_val, avg_val.
    These lists are returned as a tuple.

    :param result:List[Dict[str: Used to Store the result of each iteration.
    :param float]]: Used to Store the results of the cross validation.
    :param key_val:str: Used to Specify which key in the dictionary to use for the unpacking.
    :return: The max, min and average value of the key_val parameter.

    :doc-author: Julian M. Kleber
    """

    arr_res = np.array(unpack_run(result, key_val=key_val))

    max_val = []
    min_val = []
    avg_val = []

    for i in range(arr_res.shape[1]):
        max_val.append(get_max(arr_res[:, i].tolist()))
        min_val.append(get_min(arr_res[:, i].tolist()))
        avg_val.append(get_avg(arr_res[:, i].tolist()))
    return max_val, min_val, avg_val


def unpack_run(result_list: List[Dict[str, float]], key_val: str) -> List[List[float]]:
    """
    The unpack_run function takes a list of dictionaries and returns a list of lists.
    The input is the output from the run_experiment function, which is a list of dictionaries.
    Each dictionary contains two keys: 'params' and 'val'. The value associated with the key
    'params' is another dictionary containing all parameters used in that particular experiment.
    The value associated with the key 'val' is an array containing all values returned by
    each call to f(x).


    :param result_list:List[Dict[str: Used to Specify the type of the input parameter.
    :param float]]: Used to Specify the type of data that is expected to be returned by the function.
    :return: A list of lists, where each sublist is the result of a run.

    :doc-author: Trelent
    """

    arr_res = np.zeros((len(result_list), len(result_list[0][key_val])))

    for i, res in enumerate(result_list):
        assert len(res[key_val]) == arr_res.shape[1], str(len(res[key_val])) + str(
            arr_res.shape[1]
        )
        arr_res[i, :] = res[key_val]
    return arr_res.tolist()


def get_stacked_list(
    path_to_directory: str, num_cv: int, json_name: str
) -> List[Dict[str, float]]:
    """
    The get_stacked_list function takes in a path to a directory, the name of the column
    that we want to stack, and the number of cross-validation folds. It then returns a list
    of dictionaries that contain all of our stacked results.

    :param path_to_directory:str: Used to Specify the directory where the json files are stored.
    :param column_name:str: Used to Specify the column name of the dataframe in which we
    want to get.
    :param num_cv:int: Used to Specify the number of cross validation runs.
    :param json_name:str: Used to Specify the name of the json file that will be parsed.
    :return: A list of dictionaries, where each dictionary is the accuracy for a single cv.

    :doc-author: Julian M. Kleber
    """

    results = []

    path_to_directory = attach_slash(path_to_directory)

    for i in range(num_cv):
        logging.info(f"Attempting cv {i}")

        result = load_result_json(
            path_to_json=path_to_directory + f"CV_{i}/" + json_name
        )
        results.append(result)
        logging.info(f"parsed results for CV {i}")

    return results


def load_result_json(path_to_json: str) -> Dict[str, Any]:
    """
    The load_result_json function takes in a path to a json file and returns the contents of
    that json file as a dictionary.The function also parses the "Acc" key in the dictionary,
    which contains lists of tuples containing train and test accuracy values.
    The function then separates these tuples into two separate lists, one for train accuracy values
    and one for test accuracy values. These new lists are added to the original dictionary under
    keys "Acc_train" and "Acc_val", respectively.

    :param path_to_json:str: Used to Specify the path to the json file.
    :return: A dictionary with the following keys:.

    :doc-author: Julian M. Kleber
    """

    result = load_json_from_file(path_to_json)
    return result
